export type DflowId = string;
export type DflowNewItem<Item> = Omit<Item, "id"> & { id?: DflowId };

export type DflowNodeMetadata = {
  isAsync: boolean;
  isConstant: boolean;
};

export type DflowPinKind = "input" | "output";
export type DflowPinType =
  | "string"
  | "number"
  | "boolean"
  | "object"
  | "array"
  | "DflowId";

export type DflowRunStatus = "waiting" | "success" | "failure";

type DflowExecutionNodeInfo =
  & Pick<
    DflowSerializableNode,
    "id" | "kind" | "outputs"
  >
  & { error?: string };

export type DflowExecutionReport = {
  status: DflowRunStatus;
  start: Date;
  end?: Date;
  steps?: DflowExecutionNodeInfo[];
};

// Inspiration from https://github.com/sindresorhus/type-fest/blob/main/source/basic.d.ts
export type DflowObject = { [Key in string]?: DflowValue };
export type DflowArray = Array<DflowValue>;
export type DflowValue =
  | string
  | number
  | boolean
  | undefined
  | DflowArray
  | DflowObject;

export type DflowNodesCatalog = Record<DflowNode["kind"], typeof DflowNode>;

export type DflowSerializableItem = {
  id: DflowId;
  name?: string;
};

export type DflowSerializableNode =
  & DflowSerializableItem
  & Pick<DflowNode, "kind">
  & {
    inputs?: DflowSerializableInput[];
    outputs?: DflowSerializableOutput[];
  };

export type DflowSerializablePin = DflowSerializableItem & {
  types?: DflowPinType[];
};

export type DflowSerializableInput = DflowSerializablePin & {
  multi?: boolean;
  optional?: boolean;
};

export type DflowSerializableOutput = DflowSerializablePin & {
  data?: DflowValue;
};

export type DflowSerializablePinPath = [nodeId: DflowId, pinId: DflowId];

export type DflowSerializableEdge = DflowSerializableItem & {
  source: DflowSerializablePinPath;
  target: DflowSerializablePinPath;
};

export type DflowSerializableGraph = {
  nodes: DflowSerializableNode[];
  edges: DflowSerializableEdge[];
};

export type DflowNewEdge = DflowNewItem<DflowSerializableEdge>;
export type DflowNewInput = DflowNewItem<DflowSerializableInput>;
export type DflowNewOutput = DflowNewItem<DflowSerializableOutput>;
export type DflowNewNode =
  & Omit<DflowNewItem<DflowSerializableNode>, "inputs" | "outputs">
  & { inputs?: DflowNewInput[]; outputs?: DflowNewOutput[] };

export type DflowNodeConnection = { sourceId: DflowId; targetId: DflowId };

export type DflowNodeConstructorArg = {
  node: DflowSerializableNode;
  host: DflowHost;
};

export type DflowGraphConstructorArg = {
  nodesCatalog?: DflowNodesCatalog;
};

export type DflowHostConstructorArg = DflowGraphConstructorArg;

type DflowRunOptions = { verbose: boolean };

type DflowItemKind = DflowPinKind | "node" | "edge";

class DflowErrorItemNotFound extends Error {
  constructor(kind: DflowItemKind, id: DflowId) {
    super(`${kind} not found id=${id}`);
  }
}

const _missingString = (stringName: string) => `${stringName} must be a string`;
const _missingPin = (nodeId: DflowId, kind: DflowPinKind) =>
  `${kind} pin not found nodeId=${nodeId}`;
const _missingPinAtPosition = (
  nodeId: DflowId,
  kind: DflowPinKind,
  position: number,
) => `${_missingPin(nodeId, kind)} position=${position}`;

const _executionNodeInfo = (
  { id, kind, outputs }: DflowSerializableNode,
  error?: string,
): DflowExecutionNodeInfo => {
  const obj = {
    id,
    kind,
    outputs: outputs?.map(({ id, data, name }) => ({ id, data, name })),
  } as DflowExecutionNodeInfo;

  if (error) {
    obj.error = error;
  }

  return obj;
};

export class DflowData {
  static isArray(data: unknown): data is DflowArray {
    if (!Array.isArray(data)) return false;
    return true;
  }

  static isBoolean(data: unknown): data is boolean {
    return typeof data === "boolean";
  }

  static isDflowId(data: unknown): data is DflowId {
    return DflowData.isStringNotEmpty(data);
  }

  static isObject(data: unknown): data is DflowObject {
    if (typeof data !== "object" || !data || Array.isArray(data)) return false;
    return true;
  }

  static isNumber(data: unknown): data is number {
    return typeof data === "number" && !isNaN(data);
  }

  static isString(data: unknown): data is string {
    return typeof data === "string";
  }

  static isStringNotEmpty(data: unknown) {
    return DflowData.isString(data) && data !== "";
  }

  static validate(data: unknown, types: DflowPinType[]) {
    if (types.length === 0) {
      return true;
    }

    return types.some((pinType) => {
      switch (pinType) {
        case "array":
          return DflowData.isArray(data);
        case "boolean":
          return DflowData.isBoolean(data);
        case "number":
          return DflowData.isNumber(data);
        case "object":
          return DflowData.isObject(data);
        case "string":
          return DflowData.isString(data);
        case "DflowId":
          return DflowData.isDflowId(data);
        default:
          return false;
      }
    }, true);
  }
}

export class DflowItem {
  readonly id: DflowId;
  name?: string;

  constructor({ id, name }: DflowSerializablePin) {
    this.id = id;
    this.name = name;
  }

  toObject(): DflowSerializableItem {
    const obj = { id: this.id } as DflowSerializableItem;

    if (typeof this.name === "string") {
      obj.name = this.name;
    }

    return obj;
  }
}

export class DflowPin extends DflowItem {
  readonly kind: DflowPinKind;
  readonly types: DflowPinType[];

  static types = [
    "string",
    "number",
    "boolean",
    "object",
    "array",
    "DflowId",
  ];

  constructor(
    kind: DflowPinKind,
    { types = [], ...pin }: DflowSerializablePin,
  ) {
    super(pin);

    this.kind = kind;
    this.types = types;
  }

  get hasTypeAny() {
    return this.types.length === 0;
  }

  hasType(type: DflowPinType) {
    return this.hasTypeAny || this.types.includes(type);
  }

  toObject() {
    const obj = super.toObject() as DflowSerializablePin;

    if (this.types.length > 0) {
      obj.types = this.types;
    }

    return obj;
  }
}

export class DflowInput extends DflowPin {
  #multi?: boolean;
  #optional?: boolean;
  #source?: DflowOutput;
  #sources?: Set<DflowOutput>;

  constructor({ multi, optional, ...pin }: DflowSerializableInput) {
    super("input", pin);

    this.#multi = multi;
    this.#optional = optional;
  }

  get data(): DflowValue {
    if (this.#multi) {
      const sources = Array.from(this.#sources ?? []);
      return sources.length ? sources.map((output) => output.data) : undefined;
    } else {
      return this.#source?.data;
    }
  }

  get isConnected() {
    return this.#multi
      ? ((Array.from(this.#sources ?? [])).length > 0)
      : typeof this.#source === "undefined";
  }

  get isMulti() {
    return this.#multi;
  }

  get isOptional() {
    return this.#optional;
  }

  connectTo(pin: DflowOutput) {
    const { hasTypeAny: targetHasTypeAny, types: targetTypes } = this;
    const { types: sourceTypes } = pin;

    const canConnect = targetHasTypeAny ||
      targetTypes.some((pinType) => sourceTypes.includes(pinType));

    if (canConnect) {
      if (this.#multi) {
        if (!this.#sources) {
          this.#sources = new Set();
        }
        this.#sources.add(pin);
      } else {
        this.#source = pin;
      }
    } else {
      throw new Error(
        `mismatching pinTypes, source has types [${sourceTypes.join()}] and target has types [${targetTypes.join()}]`,
      );
    }
  }

  disconnect() {
    this.#multi ? this.#sources?.clear() : this.#source = undefined;
  }

  toObject(): DflowSerializableInput {
    return super.toObject() as DflowSerializableInput;
  }
}

export class DflowOutput extends DflowPin {
  #data?: DflowValue;

  constructor({ data, ...pin }: DflowSerializableOutput) {
    super("output", pin);

    this.#data = data;
  }

  clear() {
    this.#data = undefined;
  }

  get data(): DflowValue {
    return this.#data;
  }

  set data(data: unknown) {
    switch (true) {
      case typeof data === "undefined":
        this.clear();
        break;
      case this.hasTypeAny:
      case this.hasType("string") && DflowData.isString(data):
      case this.hasType("number") && DflowData.isNumber(data):
      case this.hasType("boolean") && DflowData.isBoolean(data):
      case this.hasType("object") && DflowData.isObject(data):
      case this.hasType("array") && DflowData.isArray(data):
      case this.hasType("DflowId") && DflowData.isDflowId(data): {
        this.#data = data as DflowValue;
        break;
      }
      default: {
        throw new Error(
          `could not set data pinTypes=${
            JSON.stringify(
              this.types,
            )
          } typeof data is ${typeof data}`,
        );
      }
    }
  }

  toObject(): DflowSerializableOutput {
    const obj = super.toObject() as DflowSerializableOutput;

    if (typeof this.#data !== "undefined") {
      obj.data = this.#data;
    }

    return obj;
  }
}

export class DflowNode extends DflowItem {
  readonly #host: DflowHost;
  readonly #inputs: Map<DflowId, DflowInput> = new Map();
  readonly #outputs: Map<DflowId, DflowOutput> = new Map();
  readonly #inputPosition: DflowId[] = [];
  readonly #outputPosition: DflowId[] = [];
  readonly kind: string;

  static kind: string;
  static isAsync?: DflowNodeMetadata["isAsync"];
  static isConstant?: DflowNodeMetadata["isConstant"];
  static inputs?: DflowNewInput[];
  static outputs?: DflowNewOutput[];

  constructor(
    {
      node: { kind, inputs = [], outputs = [], ...item },
      host,
    }: DflowNodeConstructorArg,
  ) {
    super(item);

    this.#host = host;
    this.kind = kind;

    // Inputs.
    for (const pin of inputs) {
      this.newInput(pin);
    }

    // Outputs.
    for (const pin of outputs) {
      this.newOutput(pin);
    }
  }

  static input(
    typing: DflowPinType | DflowPinType[] = [],
    rest?: Omit<DflowNewInput, "types">,
  ): DflowNewInput {
    if (typeof typing === "string") {
      return { types: [typing], ...rest };
    }
    return { types: typing, ...rest };
  }

  static output(
    typing: DflowPinType | DflowPinType[] = [],
    rest?: Omit<DflowNewOutput, "types">,
  ): DflowNewOutput {
    if (typeof typing === "string") {
      return { types: [typing], ...rest };
    }
    return { types: typing, ...rest };
  }

  get inputs() {
    return this.#inputs.values();
  }

  get outputs() {
    return this.#outputs.values();
  }

  clearOutputs() {
    for (const output of this.outputs) {
      output.clear();
    }
  }

  getInputById(id: DflowId): DflowInput {
    const item = this.#inputs.get(id);
    if (!item) {
      throw new DflowErrorItemNotFound("input", id);
    }
    return item;
  }

  input(position: number): DflowInput {
    const pinId = this.#inputPosition[position];

    if (!pinId) {
      throw new Error(_missingPinAtPosition(this.id, "input", position));
    }

    return this.getInputById(pinId);
  }

  getOutputById(id: DflowId): DflowOutput {
    const item = this.#outputs.get(id);
    if (!item) {
      throw new DflowErrorItemNotFound("output", id);
    }
    return item;
  }

  output(position: number): DflowOutput {
    const pinId = this.#outputPosition[position];

    if (!pinId) {
      throw new Error(_missingPinAtPosition(this.id, "output", position));
    }

    return this.getOutputById(pinId);
  }

  deleteInput(pinId: DflowId) {
    this.#host.deleteEdgesConnectedToPin([this.id, pinId]);
    this.#inputs.delete(pinId);
    this.#inputPosition.splice(this.#inputPosition.indexOf(pinId), 1);
  }

  deleteOutput(pinId: DflowId) {
    this.#host.deleteEdgesConnectedToPin([this.id, pinId]);
    this.#outputs.delete(pinId);
    this.#outputPosition.splice(this.#outputPosition.indexOf(pinId), 1);
  }

  newInput(obj: DflowNewInput): DflowInput {
    const numInputs = this.#inputs.size;

    const generateInputId = (i = numInputs): DflowId => {
      const id = `i${i}`;
      return this.#inputs.has(id) ? generateInputId(i + 1) : id;
    };

    const id = DflowData.isDflowId(obj.id) ? obj.id : generateInputId();
    const pin = new DflowInput({ ...obj, id });
    this.#inputs.set(id, pin);
    this.#inputPosition.push(id);
    return pin;
  }

  newOutput(obj: DflowNewOutput): DflowOutput {
    const numOutputs = this.#outputs.size;

    const generateOutputId = (i = numOutputs): DflowId => {
      const id = `o${i}`;
      return this.#outputs.has(id) ? generateOutputId(i + 1) : id;
    };

    const id = DflowData.isDflowId(obj.id) ? obj.id : generateOutputId();
    const pin = new DflowOutput({ ...obj, id });
    this.#outputs.set(id, pin);
    this.#outputPosition.push(id);
    return pin;
  }

  run(): void {}

  toObject(): DflowSerializableNode {
    const obj = {
      ...super.toObject(),
      kind: this.kind,
    } as DflowSerializableNode;

    const inputs = [];
    const outputs = [];

    for (const input of this.inputs) {
      inputs.push(input.toObject());
    }
    if (inputs.length > 0) {
      obj.inputs = inputs;
    }

    for (const output of this.outputs) {
      outputs.push(output.toObject());
    }
    if (outputs.length > 0) {
      obj.outputs = outputs;
    }

    return obj;
  }
}

export class DflowEdge extends DflowItem {
  readonly source: DflowSerializablePinPath;
  readonly target: DflowSerializablePinPath;

  constructor({ source, target, ...item }: DflowSerializableEdge) {
    super(item);

    this.source = source;
    this.target = target;
  }

  toObject(): DflowSerializableEdge {
    return {
      ...super.toObject(),
      source: this.source,
      target: this.target,
    };
  }
}

export class DflowGraph {
  readonly nodesCatalog: DflowNodesCatalog;
  readonly nodes: Map<DflowId, DflowNode> = new Map();
  readonly edges: Map<DflowId, DflowEdge> = new Map();
  runOptions: DflowRunOptions = { verbose: false };
  runStatus: DflowRunStatus | null = null;
  executionReport: DflowExecutionReport | null = null;

  constructor({ nodesCatalog = {} }: DflowGraphConstructorArg = {}) {
    this.nodesCatalog = { ...nodesCatalog, ...coreNodesCatalog };
  }

  static childrenOfNodeId(
    nodeId: DflowId,
    nodeConnections: { sourceId: DflowId; targetId: DflowId }[],
  ) {
    return nodeConnections
      .filter(({ sourceId }) => nodeId === sourceId)
      .map(({ targetId }) => targetId);
  }

  static parentsOfNodeId(
    nodeId: DflowId,
    nodeConnections: { sourceId: DflowId; targetId: DflowId }[],
  ) {
    return nodeConnections
      .filter(({ targetId }) => nodeId === targetId)
      .map(({ sourceId }) => sourceId);
  }

  static levelOfNodeId(
    nodeId: DflowId,
    nodeConnections: DflowNodeConnection[],
  ) {
    const parentsNodeIds = DflowGraph.parentsOfNodeId(nodeId, nodeConnections);
    // 1. A node with no parent as level zero.
    if (parentsNodeIds.length === 0) {
      return 0;
    }

    // 2. Otherwise its level is the max level of its parents plus one.
    let maxLevel = 0;
    for (const parentNodeId of parentsNodeIds) {
      const level = DflowGraph.levelOfNodeId(parentNodeId, nodeConnections);
      maxLevel = Math.max(level, maxLevel);
    }
    return maxLevel + 1;
  }

  static ancestorsOfNodeId(
    nodeId: DflowId,
    nodeConnections: DflowNodeConnection[],
  ): DflowId[] {
    const parentsNodeIds = DflowGraph.parentsOfNodeId(nodeId, nodeConnections);
    if (parentsNodeIds.length === 0) {
      return [];
    } else {
      return parentsNodeIds.reduce<DflowId[]>(
        (accumulator, parentNodeId, index, array) => {
          const ancestors = DflowGraph.ancestorsOfNodeId(
            parentNodeId,
            nodeConnections,
          );

          const result = accumulator.concat(ancestors);

          // On last iteration, remove duplicates
          return index === array.length - 1
            ? Array.from(new Set(array.concat(result)))
            : result;
        },
        [],
      );
    }
  }

  static sortNodesByLevel(
    nodeIds: DflowId[],
    nodeConnections: DflowNodeConnection[],
  ): DflowId[] {
    const levelOf: Record<DflowId, number> = {};

    for (const nodeId of nodeIds) {
      levelOf[nodeId] = DflowGraph.levelOfNodeId(nodeId, nodeConnections);
    }

    return nodeIds.slice().sort((a, b) => (levelOf[a] <= levelOf[b] ? -1 : 1));
  }

  get nodeConnections(): DflowNodeConnection[] {
    return [...this.edges.values()].map((edge) => ({
      sourceId: edge.source[0],
      targetId: edge.target[0],
    }));
  }

  get nodeIdsInsideFunctions(): DflowId[] {
    const ancestorsOfReturnNodes = [];

    // Find all "return" nodes and get their ancestors.
    for (const node of [...this.nodes.values()]) {
      if (node.kind === "return") {
        ancestorsOfReturnNodes.push(
          DflowGraph.ancestorsOfNodeId(node.id, this.nodeConnections),
        );
      }
    }

    // Flatten and deduplicate results.
    return Array.from(new Set(ancestorsOfReturnNodes.flat()));
  }

  async run() {
    const { verbose } = this.runOptions;

    // Set runStatus to waiting if there was some unhandled error in a previous run.
    this.runStatus = "waiting";

    this.executionReport = {
      status: this.runStatus,
      start: new Date(),
    };

    if (verbose) {
      this.executionReport.steps = [];
    }

    // Get nodeIds
    // 1. filtered by nodes inside functions
    // 2. sorted by graph hierarchy
    const nodeIdsExcluded = this.nodeIdsInsideFunctions;
    const nodeIds = DflowGraph.sortNodesByLevel(
      [...this.nodes.keys()].filter((nodeId) =>
        !nodeIdsExcluded.includes(nodeId)
      ),
      this.nodeConnections,
    );

    // Two nested loops:
    //   1. NODES_LOOP
    //   2. INPUTS_LOOP
    //
    // For every node loop over their inputs and validate data.

    // 1. NODES_LOOP
    // /////////////
    NODES_LOOP:
    for (const nodeId of nodeIds) {
      const node = this.nodes.get(nodeId) as DflowNode;

      const NodeClass = this.nodesCatalog[node.kind] ?? DflowNodeUnknown;
      const { isAsync, isConstant } = NodeClass;

      try {
        if (!isConstant) {
          let someInputIsNotValid = false;

          // 2. INPUTS_LOOP
          // //////////////
          INPUTS_LOOP:
          for (const { id, data, types, isOptional } of node.inputs) {
            // Ignore optional inputs.
            if (isOptional && typeof data === "undefined") {
              continue INPUTS_LOOP;
            }

            // Validate input data.
            if (!DflowData.validate(data, types)) {
              someInputIsNotValid = true;
              if (verbose) {
                this.executionReport.steps?.push(
                  _executionNodeInfo(
                    node.toObject(),
                    `invalid input data nodeId=${nodeId} inputId=${id} data=${data}`,
                  ),
                );
              }
              break INPUTS_LOOP;
            }
          }

          if (someInputIsNotValid) {
            for (const output of node.outputs) {
              output.clear();
            }
            continue NODES_LOOP;
          }

          if (isAsync) {
            await node.run();
          } else {
            node.run();
          }
        }

        if (verbose) {
          this.executionReport.steps?.push(_executionNodeInfo(node.toObject()));
        }
      } catch (error) {
        console.error(error);
        this.runStatus = "failure";
      }
    }

    // Set runStatus to success if there was no error.
    if (this.runStatus === "waiting") {
      this.runStatus = "success";
    }

    this.executionReport.status = this.runStatus;
    this.executionReport.end = new Date();
  }

  toObject(): DflowSerializableGraph {
    const obj = {
      nodes: [],
      edges: [],
    } as DflowSerializableGraph;

    for (const node of this.nodes.values()) {
      obj.nodes.push(node.toObject());
    }
    for (const edge of this.edges.values()) {
      obj.edges.push(edge.toObject());
    }

    return obj;
  }
}

export class DflowHost {
  readonly #graph: DflowGraph;
  readonly context: Record<string, unknown>;

  constructor(arg?: DflowHostConstructorArg) {
    this.#graph = new DflowGraph(arg);
    this.context = {};
  }

  get executionReport() {
    return this.#graph.executionReport;
  }

  get edges() {
    return Array.from(this.#graph.edges.values());
  }

  get nodes() {
    return Array.from(this.#graph.nodes.values());
  }

  get nodesCatalog(): DflowNodesCatalog {
    return this.#graph.nodesCatalog;
  }

  get runStatusIsSuccess() {
    return this.#graph.runStatus === "success";
  }

  get runStatusIsWaiting() {
    return this.#graph.runStatus === "waiting";
  }

  get runStatusIsFailure() {
    return this.#graph.runStatus === "failure";
  }

  set verbose(option: DflowRunOptions["verbose"]) {
    this.#graph.runOptions.verbose = option;
  }

  clearGraph() {
    this.#graph.nodes.clear();
    this.#graph.edges.clear();
  }

  connect(sourceNode: DflowNode, sourcePosition = 0) {
    return {
      to: (targetNode: DflowNode, targetPosition = 0) => {
        const sourcePin = sourceNode.output(sourcePosition);
        const targetPin = targetNode.input(targetPosition);

        this.newEdge({
          source: [sourceNode.id, sourcePin.id],
          target: [targetNode.id, targetPin.id],
        });
      },
    };
  }

  deleteEdge(edgeId: DflowId) {
    if (typeof edgeId !== "string") {
      throw new TypeError(_missingString("edgeId"));
    }

    const edge = this.getEdgeById(edgeId);
    if (!edge) return;

    // 1. Cleanup target pin.
    const [targetNodeId, targetPinId] = edge.target;
    const targetNode = this.getNodeById(targetNodeId);
    const targetPin = targetNode.getInputById(targetPinId);
    targetPin.disconnect();

    // 2. Delete edge.
    this.#graph.edges.delete(edgeId);
  }

  deleteNode(nodeId: DflowId) {
    if (typeof nodeId !== "string") {
      throw new TypeError(_missingString("nodeId"));
    }

    const node = this.getNodeById(nodeId);

    if (node) {
      // 1. Delete all edges connected to node.
      for (const edge of this.edges) {
        const {
          source: [sourceNodeId],
          target: [targetNodeId],
        } = edge;
        if (sourceNodeId === node.id || targetNodeId === node.id) {
          this.deleteEdge(edge.id);
        }
      }

      // 2. Delete node.
      this.#graph.nodes.delete(nodeId);
    } else {
      throw new Error(`DflowNode not found, id=${nodeId}`);
    }
  }

  deleteEdgesConnectedToPin([nodeId, pinId]: DflowSerializablePinPath) {
    for (const edge of this.edges) {
      const [sourceNodeId, sourcePinId] = edge.source;
      const [targetNodeId, targetPinId] = edge.target;

      if (
        (sourceNodeId === nodeId && sourcePinId === pinId) ||
        (targetNodeId === nodeId && targetPinId === pinId)
      ) {
        this.deleteEdge(edge.id);
      }
    }
  }

  executeFunction(functionId: DflowId, args: DflowArray) {
    const { verbose } = this.#graph.runOptions;

    // Get all return nodes connected to function node.
    const nodeConnections = this.#graph.nodeConnections;
    const childrenNodeIds = DflowGraph.childrenOfNodeId(
      functionId,
      nodeConnections,
    );
    const returnNodeIds = [];
    for (const childrenNodeId of childrenNodeIds) {
      const node = this.getNodeById(childrenNodeId);
      if (node.kind === DflowNodeReturn.kind) {
        returnNodeIds.push(node.id);
      }
    }

    // Get all nodes inside function.
    const nodeIdsInsideFunction = returnNodeIds.reduce<DflowId[]>(
      (accumulator, returnNodeId, index, array) => {
        const ancestors = DflowGraph.ancestorsOfNodeId(
          returnNodeId,
          nodeConnections,
        );

        const result = accumulator.concat(ancestors);

        // On last iteration, remove duplicates
        return index === array.length ? Array.from(new Set(result)) : result;
      },
      [],
    );

    // 1. get nodeIds sorted by graph hierarchy
    // 2. if it is an argument node, inject input data
    // 3. if if is a return node, output data
    // 4. otherwise run node
    const nodeIds = DflowGraph.sortNodesByLevel(
      [...returnNodeIds, ...nodeIdsInsideFunction],
      nodeConnections,
    );
    for (const nodeId of nodeIds) {
      const node = this.getNodeById(nodeId) as DflowNode;

      const NodeClass = this.nodesCatalog[node.kind] ?? DflowNodeUnknown;
      const { isAsync, isConstant } = NodeClass;

      try {
        switch (node.kind) {
          case DflowNodeArgument.kind: {
            const position = node.input(0).data;
            // Argument position default to 0, must be >= 0.
            const index = typeof position === "number" && !isNaN(position)
              ? Math.max(
                position,
                0,
              )
              : 0;
            node.output(0).data = args[index];
            break;
          }
          case DflowNodeReturn.kind: {
            return node.input(1).data;
          }
          default: {
            // Notice that executeFunction cannot execute async functions.
            if (!isConstant && !isAsync) {
              node.run();
            }

            if (verbose) {
              this.executionReport?.steps?.push(
                _executionNodeInfo(node.toObject()),
              );
            }
          }
        }
      } catch (error) {
        console.error(error);
      }
    }
  }

  getEdgeById(id: DflowId): DflowEdge | undefined {
    const item = this.#graph.edges.get(id);
    if (!item) {
      throw new DflowErrorItemNotFound("edge", id);
    }
    return item;
  }

  getNodeById(id: DflowId): DflowNode {
    const item = this.#graph.nodes.get(id);
    if (!item) {
      throw new DflowErrorItemNotFound("node", id);
    }
    return item;
  }

  newNode(obj: DflowNewNode): DflowNode {
    const numNodes = this.#graph.nodes.size;

    const generateNodeId = (i = numNodes): DflowId => {
      const id = `n${i}`;
      return this.#graph.nodes.has(id) ? generateNodeId(i + 1) : id;
    };

    const NodeClass = this.nodesCatalog[obj.kind] ?? DflowNodeUnknown;

    const id = DflowData.isDflowId(obj.id)
      ? (obj.id as DflowId)
      : generateNodeId();

    const inputs = NodeClass.inputs?.map((pin, i) => {
      const objPin = obj.inputs?.[i] ?? {} as Partial<DflowNewInput>;
      const id = DflowData.isDflowId(objPin?.id) ? objPin.id : `i${i}`;
      return {
        id,
        ...objPin,
        ...pin,
      };
    }) ?? [];

    const outputs = NodeClass.outputs?.map((pin, i) => {
      const objPin = obj.outputs?.[i] ?? {} as Partial<DflowNewOutput>;
      const id = DflowData.isDflowId(objPin?.id) ? objPin.id : `o${i}`;
      return {
        id,
        ...objPin,
        ...pin,
      };
    }) ?? [];

    const node = new NodeClass({
      node: { ...obj, id, inputs, outputs },
      host: this,
    });

    this.#graph.nodes.set(node.id, node);

    return node;
  }

  newEdge(obj: DflowNewEdge): DflowEdge {
    const numEdges = this.#graph.edges.size;

    const generateEdgeId = (i = numEdges): DflowId => {
      const id = `e${i}`;
      return this.#graph.edges.has(id) ? generateEdgeId(i + 1) : id;
    };

    const id = DflowData.isDflowId(obj.id)
      ? (obj.id as DflowId)
      : generateEdgeId();

    const edge = new DflowEdge({ ...obj, id });

    this.#graph.edges.set(edge.id, edge);

    const [sourceNodeId, sourcePinId] = edge.source;
    const [targetNodeId, targetPinId] = edge.target;

    const sourceNode = this.getNodeById(sourceNodeId);
    const targetNode = this.getNodeById(targetNodeId);
    const sourcePin = sourceNode.getOutputById(sourcePinId);
    const targetPin = targetNode.getInputById(targetPinId);

    targetPin.connectTo(sourcePin);

    return edge;
  }

  newInput(nodeId: DflowId, obj: DflowNewInput): DflowInput {
    const node = this.getNodeById(nodeId);
    return node.newInput(obj);
  }

  newOutput(nodeId: DflowId, obj: DflowNewOutput): DflowOutput {
    const node = this.getNodeById(nodeId);
    return node.newOutput(obj);
  }

  toObject() {
    return this.#graph.toObject();
  }

  async run() {
    await this.#graph.run();
  }
}

// Core nodes.

const { input, output } = DflowNode;

class DflowNodeArgument extends DflowNode {
  static kind = "argument";
  static isConstant = true;
  static inputs = [input("number", { name: "position", optional: true })];
  static outputs = [output()];
}

class DflowNodeArray extends DflowNode {
  static kind = "array";
  static inputs = [input()];
  static outputs = [output("array")];
  run() {
    const data = this.input(0).data;
    if (DflowData.isArray(data)) {
      this.output(0).data = data;
    } else {
      this.output(0).clear();
    }
  }
}

class DflowNodeBoolean extends DflowNode {
  static kind = "boolean";
  static inputs = [input()];
  static outputs = [output("boolean")];
  run() {
    const data = this.input(0).data;
    if (DflowData.isBoolean(data)) {
      this.output(0).data = data;
    } else {
      this.output(0).clear();
    }
  }
}

class DflowNodeData extends DflowNode {
  static kind = "data";
  static isConstant = true;
  static outputs = [output()];
  constructor({ node: { outputs, ...node }, host }: DflowNodeConstructorArg) {
    super({
      node: {
        ...node,
        outputs: outputs?.map((output) => ({
          ...output,
          types: DflowNodeData.inferDflowPinTypes(output.data),
        })),
      },
      host,
    });
  }
  static inferDflowPinTypes(data: DflowValue): DflowPinType[] {
    switch (true) {
      case DflowData.isArray(data):
        return ["array"];
      case DflowData.isBoolean(data):
        return ["boolean"];
      case DflowData.isNumber(data):
        return ["number"];
      case DflowData.isString(data):
        return ["string"];
      case DflowData.isObject(data):
        return ["object"];
      default:
        return [];
    }
  }
}

class DflowNodeFunction extends DflowNode {
  static kind = "function";
  static isConstant = true;
  static outputs = [output("DflowId", { name: "id" })];
  constructor(arg: DflowNodeConstructorArg) {
    super(arg);
    this.output(0).data = this.id;
  }
}

class DflowNodeIsUndefined extends DflowNode {
  static kind = "isUndefined";
  static inputs = [input()];
  static outputs = [output("boolean")];
  run() {
    this.output(0).data = typeof this.input(0).data === "undefined";
  }
}

class DflowNodeNumber extends DflowNode {
  static kind = "number";
  static inputs = [input()];
  static outputs = [output("number")];
  run() {
    const data = this.input(0).data;
    if (DflowData.isNumber(data)) {
      this.output(0).data = data;
    } else {
      this.output(0).clear();
    }
  }
}

class DflowNodeObject extends DflowNode {
  static kind = "object";
  static inputs = [input()];
  static outputs = [output("object")];
  run() {
    const data = this.input(0).data;
    if (DflowData.isObject(data)) {
      this.output(0).data = data;
    } else {
      this.output(0).clear();
    }
  }
}

class DflowNodeReturn extends DflowNode {
  static kind = "return";
  static isConstant = true;
  static inputs = [
    input("DflowId", { name: "functionId" }),
    input([], { name: "value" }),
  ];
}

class DflowNodeString extends DflowNode {
  static kind = "string";
  static inputs = [input()];
  static outputs = [output("string")];
  run() {
    const data = this.input(0).data;
    if (DflowData.isString(data)) {
      this.output(0).data = data;
    } else {
      this.output(0).clear();
    }
  }
}

// The "Unknown" node is not inclued in core nodes catalog.
class DflowNodeUnknown extends DflowNode {}

const coreNodesCatalog: DflowNodesCatalog = {
  [DflowNodeArgument.kind]: DflowNodeArgument,
  [DflowNodeArray.kind]: DflowNodeArray,
  [DflowNodeBoolean.kind]: DflowNodeBoolean,
  [DflowNodeData.kind]: DflowNodeData,
  [DflowNodeIsUndefined.kind]: DflowNodeIsUndefined,
  [DflowNodeNumber.kind]: DflowNodeNumber,
  [DflowNodeObject.kind]: DflowNodeObject,
  [DflowNodeFunction.kind]: DflowNodeFunction,
  [DflowNodeString.kind]: DflowNodeString,
  [DflowNodeReturn.kind]: DflowNodeReturn,
};
